open List

(* A Node has a key, value and list of children *)
type ('a, 'b) trie = Empty | Node of 'a list * 'b * ('a, 'b) trie list

let rec starts_with l s =
    (length s) <= (length l) && match s with
    | [] -> true
    | h::t -> h = (hd l) && starts_with (tl l) t
;;

(* Insert a new (key, value) pair into a trie *)
let rec insert trie key value =
    match trie with
    | Empty -> Node (key, value, []) (* Root *)
    | Node (k, v, children) ->
        if k = key then
            raise (Failure "Inserted key already exists in trie")
        else if starts_with key k then
            (* Inserted key should be in this node because it starts with
             * the node's key *)
            let rec walk_nodes = function
                | [] -> Empty
                | node::tail ->
                    match node with
                    | Empty -> Empty
                    | Node (k, _, _) ->
                        if starts_with key k then
                            Node (k, v, (insert node key value)::children)
                        else
                            walk_nodes tail
            in
            (* First, check if the new pair should be inserted in one of the
             * children. If not, prepend it to the children list while moving
             * all children that are should be inside the new node *)
            match walk_nodes children with
            | Empty ->
                let matches_key = function
                    | Empty -> false
                    | Node (k, _, _) -> starts_with k key
                in
                (* 'move' are the children that are moved to be the children of
                 * the inserted node, 'siblings' are the children that remain
                 * in the current matched node *)
                let move, siblings = partition matches_key children in
                Node (k, v, (Node (key, value, move))::siblings)
            | node -> node
        else
            raise (Failure "Inserted key does not start with node key")
;;

(* Look up a value associated with a key *)
let rec lookup trie key =
    match trie with
    | Empty -> None (* Trie is empty, so no result *)
    | Node (k, value, children) ->
        if k = key then
            (* Keys are equal, return value *)
            Some value
        else if starts_with key k then
            (* The node's key prefixes the given key, so the value can only
             * be in one of the children. *)
            let rec walk_nodes = function
                | [] -> None
                | child::rest -> match lookup child key with
                    | None -> walk_nodes rest
                    | result -> result
            in
            walk_nodes children
        else
            (* Not found *)
            None
;;

(* Find all (key, value) pairs whose key start with a given prefix *)
let rec matches trie key =
    match trie with
    | Empty -> []
    | Node (k, value, children) ->
        let rec match_nodes = function
            | [] -> []
            | node::rest -> (matches node key) @ (match_nodes rest)
        in
        (if starts_with k key then [(k, value)] else [])
        @ match_nodes children
;;

(*let a = Node ([1;2;3], "a", []);;
let b = Node ([1;3], "b", [a]);;
let c = Node ([3;4], "c", [b]);;*)
